"""
Flight controller connection progress UI module.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2026 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import sys
import tkinter as tk
from logging import error as logging_error
from tkinter import ttk
from types import TracebackType
from typing import Literal, Optional

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.frontend_tkinter_progress_window import ProgressWindow

if sys.version_info >= (3, 11):
    from typing import Self
else:  # Python 3.9 and 3.10
    from typing_extensions import Self

PROGRESS_FC_INIT_COMPLETE = 20


class FlightControllerConnectionProgress:
    """
    UI class for flight controller connection progress.

    Manages the progress window and temporary Tk root for connection UI.
    """

    def __init__(self) -> None:
        # Needed because ProgressWindow needs a master and there is no application root window at this point
        self.temp_root: tk.Tk = tk.Tk(className="ArduPilotMethodicConfigurator")
        self.temp_root.withdraw()  # Hide the temporary root window, we only want to show the progress window

        # Set the theme to 'alt'
        style = ttk.Style()
        style.theme_use("alt")

        self.progress_window: ProgressWindow = ProgressWindow(
            self.temp_root,
            _("Initializing Flight Controller"),
            _("Starting initialization..."),
            only_show_when_update_progress_called=True,
        )
        # Note: ProgressWindow handles centering itself when parent is a minimal temp root

    def update_init_progress_bar(self, value: int, max_value: int) -> None:
        """
        Update progress bar for initialization phase.

        Maps input range [0, 100] to output range [0, PROGRESS_FC_INIT_COMPLETE].

        Args:
            value: Current progress value (0-100)
            max_value: Maximum progress value (typically 100)

        """
        if max_value <= 0:
            logging_error(_("Invalid max_value in update_init_progress_bar progress update: %d"), max_value)
            return
        self.progress_window.update_progress_bar(value * PROGRESS_FC_INIT_COMPLETE // 100, max_value)

    def update_connect_progress_bar(self, value: int, max_value: int) -> None:
        """
        Update progress bar for connection phase.

        Maps input range [0, 100] to output range [PROGRESS_FC_INIT_COMPLETE, 100].

        Args:
            value: Current progress value (0-100)
            max_value: Maximum progress value (typically 100)

        """
        if max_value <= 0:
            logging_error(_("Invalid max_value in update_connect_progress_bar progress update: %d"), max_value)
            return
        self.progress_window.update_progress_bar(
            PROGRESS_FC_INIT_COMPLETE + value * (100 - PROGRESS_FC_INIT_COMPLETE) // 100, max_value
        )

    def destroy(self) -> None:
        """Clean up resources - destroys progress window and temporary root."""
        if self.progress_window:
            self.progress_window.destroy()  # Destroy child first
        if self.temp_root:
            self.temp_root.destroy()  # Then destroy parent

    def __enter__(self) -> Self:
        """Enter context manager - returns self for use in 'with' statement."""
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Literal[False]:
        """
        Exit context manager - ensures cleanup happens.

        Args:
            exc_type: Exception type if an exception occurred, None otherwise
            exc_val: Exception value if an exception occurred, None otherwise
            exc_tb: Exception traceback if an exception occurred, None otherwise

        Returns:
            False to propagate any exception that occurred

        """
        self.destroy()
        return False  # Don't suppress exceptions



#!/usr/bin/env python3

"""
GUI to select the directory to store the vehicle configuration files.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2026 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import tkinter as tk
from argparse import ArgumentParser, Namespace
from logging import basicConfig as logging_basicConfig
from logging import debug as logging_debug
from logging import error as logging_error
from logging import getLevelName as logging_getLevelName
from logging import warning as logging_warning
from os import path as os_path
from sys import exit as sys_exit
from tkinter import messagebox, ttk

from ardupilot_methodic_configurator import _, __version__
from ardupilot_methodic_configurator.backend_filesystem import LocalFilesystem
from ardupilot_methodic_configurator.common_arguments import add_common_arguments
from ardupilot_methodic_configurator.data_model_vehicle_project import VehicleProjectManager
from ardupilot_methodic_configurator.data_model_vehicle_project_opener import VehicleProjectOpenError
from ardupilot_methodic_configurator.frontend_tkinter_base_window import BaseWindow
from ardupilot_methodic_configurator.frontend_tkinter_directory_selection import (
    VehicleDirectorySelectionWidgets,
)
from ardupilot_methodic_configurator.frontend_tkinter_project_creator import VehicleProjectCreatorWindow
from ardupilot_methodic_configurator.frontend_tkinter_show import show_tooltip


class VehicleProjectOpenerWindow(BaseWindow):
    """
    A window for selecting a vehicle directory with intermediate parameter files.

    This class extends the BaseWindow class to provide a graphical user interface
    for selecting a vehicle directory that contains intermediate parameter files
    for ArduPilot. It allows the user to choose between creating a new vehicle
    configuration directory based on an existing template or using an existing
    vehicle configuration directory.
    """

    # pylint: disable=duplicate-code
    def __init__(self, project_manager: VehicleProjectManager) -> None:
        super().__init__()
        self.project_manager = project_manager
        self.root.title(
            _("Amilcar Lucas's - ArduPilot methodic configurator ")
            + __version__
            + _(" - Select vehicle configuration directory")
        )

        self.root.geometry("600x450")  # Set the window size

        # Explain why we are here
        introduction_text = self.project_manager.get_introduction_message()
        introduction_label = ttk.Label(
            self.main_frame,
            anchor=tk.CENTER,
            justify=tk.CENTER,
            text=introduction_text + _("\nChoose one of the following three options:"),
        )
        introduction_label.pack(expand=False, fill=tk.X, padx=6, pady=6)
        _template_dir, _new_base_dir, vehicle_dir = self.project_manager.get_recently_used_dirs()
        logging_debug("vehicle_dir: %s", vehicle_dir)  # this string is intentionally left untranslated
        self.create_option1_widgets()
        self.create_option2_widgets(vehicle_dir)
        self.create_option3_widgets()

        # Bind the close_connection_and_quit function to the window close event
        self.root.protocol("WM_DELETE_WINDOW", self.close_and_quit)

    def close_and_quit(self) -> None:
        sys_exit(0)

    def create_option1_widgets(self) -> None:
        # Option 1 - Create a new vehicle configuration directory based on an existing template
        option1_label = ttk.Label(text=_("New vehicle"), style="Bold.TLabel")
        option1_label_frame = ttk.LabelFrame(self.main_frame, labelwidget=option1_label)
        option1_label_frame.pack(expand=True, fill=tk.X, padx=6, pady=6)

        create_vehicle_directory_from_template_button = ttk.Button(
            option1_label_frame,
            text=_("Create a vehicle configuration directory from template"),
            command=self.create_new_vehicle_from_template,
        )
        create_vehicle_directory_from_template_button.pack(expand=False, fill=tk.X, padx=20, pady=5, anchor=tk.CENTER)
        show_tooltip(
            create_vehicle_directory_from_template_button,
            _("Create a new vehicle configuration directory, choose this option when using the software for the first time"),
        )

    # pylint: enable=duplicate-code

    def create_option2_widgets(self, initial_dir: str) -> None:
        # Option 2 - Use an existing vehicle configuration directory
        option2_label = ttk.Label(text=_("Open vehicle"), style="Bold.TLabel")
        option2_label_frame = ttk.LabelFrame(self.main_frame, labelwidget=option2_label)
        option2_label_frame.pack(expand=True, fill=tk.X, padx=6, pady=6)
        option2_label = ttk.Label(
            option2_label_frame,
            anchor=tk.CENTER,
            justify=tk.CENTER,
            text=_(
                "Use an existing vehicle configuration directory with\n"
                "intermediate parameter files, apm.pdef.xml and vehicle_components.json"
            ),
        )
        option2_label.pack(expand=False, fill=tk.X, padx=6)

        def on_vehicle_directory_selected(directory: str) -> None:
            self.project_manager.open_vehicle_directory(directory)

        self.connection_selection_widgets = VehicleDirectorySelectionWidgets(
            self,
            option2_label_frame,
            initial_dir,
            destroy_parent_on_open=True,
            on_select_directory_callback=on_vehicle_directory_selected,
        )
        self.connection_selection_widgets.container_frame.pack(expand=True, fill=tk.X, padx=3, pady=5, anchor=tk.NW)

    def create_option3_widgets(self) -> None:
        # Option 3 - Open the last used vehicle configuration directory
        option3_label = ttk.Label(text=_("Re-Open vehicle"), style="Bold.TLabel")
        option3_label_frame = ttk.LabelFrame(self.main_frame, labelwidget=option3_label)
        option3_label_frame.pack(expand=True, fill=tk.X, padx=6, pady=6)

        # Get recent vehicle directories for the combobox
        recent_vehicle_history = self.project_manager.get_recent_vehicle_dirs()

        # Create a label for the combobox
        recent_dirs_label = ttk.Label(option3_label_frame, text=_("Recently opened vehicle directories:"))
        recent_dirs_label.pack(side=tk.TOP, anchor=tk.NW, padx=3, pady=(5, 0))
        show_tooltip(recent_dirs_label, _("Select from recently opened vehicle configuration directories"))

        # Create a frame for the combobox
        combobox_frame = ttk.Frame(option3_label_frame)
        combobox_frame.pack(expand=False, fill=tk.X, padx=3, pady=(0, 5), anchor=tk.NW)

        # Create a StringVar to track the selected value
        self.recent_dir_var = tk.StringVar(value=recent_vehicle_history[0] if recent_vehicle_history else "")

        # Create the readonly combobox
        self.recent_dirs_combobox = ttk.Combobox(
            combobox_frame,
            textvariable=self.recent_dir_var,
            values=recent_vehicle_history,
            state="readonly",
        )
        self.recent_dirs_combobox.pack(side=tk.LEFT, fill=tk.X, expand=True, pady=(4, 0))
        show_tooltip(self.recent_dirs_combobox, _("Select from recently opened vehicle configuration directories"))

        # Check if there are any recent directories available
        has_recent_dirs = bool(recent_vehicle_history)
        button_state = tk.NORMAL if has_recent_dirs else tk.DISABLED

        self.open_recent_vehicle_directory_button = ttk.Button(
            option3_label_frame,
            text=_("Open selected vehicle configuration directory"),
            command=self.open_selected_recent_vehicle_directory,
            state=button_state,
        )
        self.open_recent_vehicle_directory_button.pack(expand=False, fill=tk.X, padx=20, pady=5, anchor=tk.CENTER)
        show_tooltip(
            self.open_recent_vehicle_directory_button,
            _("Open the selected vehicle configuration directory for configuring and tuning the vehicle"),
        )

    def open_selected_recent_vehicle_directory(self) -> None:
        """Open the vehicle directory selected in the recent directories combobox."""
        selected_dir = self.recent_dir_var.get()
        if not selected_dir:
            messagebox.showerror(
                _("No Directory Selected"), _("Please select a vehicle configuration directory from the list.")
            )
            return

        # Check if the directory still exists and is accessible
        try:
            if not os_path.exists(selected_dir):
                messagebox.showerror(
                    _("Directory Not Found"),
                    _("The selected directory no longer exists:\n\n{}\n\nIt may have been moved or deleted.").format(
                        selected_dir
                    ),
                )
                return

            if not os_path.isdir(selected_dir):
                messagebox.showerror(
                    _("Invalid Directory"),
                    _("The selected path is not a directory:\n\n{}").format(selected_dir),
                )
                return
        except PermissionError:
            messagebox.showerror(
                _("Permission Denied"),
                _(
                    "You do not have permission to access the selected directory:\n\n{}\n\n"
                    "Please check your file system permissions."
                ).format(selected_dir),
            )
            return
        except OSError as e:
            messagebox.showerror(
                _("File System Error"),
                _("An error occurred while accessing the directory:\n\n{}\n\nError: {}").format(selected_dir, str(e)),
            )
            return

        self.open_last_vehicle_directory(selected_dir)

    def create_new_vehicle_from_template(self) -> None:
        # close this window and open a VehicleProjectCreatorWindow instance
        self.root.destroy()
        VehicleProjectCreatorWindow(self.project_manager)

    def open_last_vehicle_directory(self, last_vehicle_dir: str) -> None:
        # Attempt to open the last opened vehicle configuration directory
        try:
            self.project_manager.open_last_vehicle_directory(last_vehicle_dir)
            self.root.destroy()
        except VehicleProjectOpenError as e:
            messagebox.showerror(e.title, e.message)


# pylint: disable=duplicate-code
def argument_parser() -> Namespace:  # pragma: no cover
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.

    """
    parser = ArgumentParser(
        description=_(
            "This main is for testing and development only. "
            "Usually, the VehicleProjectOpenerWindow is called from another script"
        )
    )
    parser = LocalFilesystem.add_argparse_arguments(parser)
    return add_common_arguments(parser).parse_args()


def main() -> None:  # pragma: no cover
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format="%(asctime)s - %(levelname)s - %(message)s")

    logging_warning(
        _(
            "This main is for testing and development only, usually the VehicleProjectOpenerWindow is"
            " called from another script"
        )
    )

    local_filesystem = LocalFilesystem(
        args.vehicle_dir, args.vehicle_type, "", args.allow_editing_template_files, args.save_component_to_system_templates
    )

    # Create project manager with the local filesystem
    project_manager = VehicleProjectManager(local_filesystem)

    # Get the list of intermediate parameter files files that will be processed sequentially
    files = project_manager.get_file_parameters_list()

    if not files:
        logging_error(_("No intermediate parameter files found in %s."), args.vehicle_dir)

    window = VehicleProjectOpenerWindow(project_manager)
    window.root.mainloop()


# pylint: enable=duplicate-code


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Recent items history list manager.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2026 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from typing import Any, Callable, Optional

from ardupilot_methodic_configurator import _


class RecentItemsHistoryList:
    """
    Helper class to manage most-recent-first (LIFO) history lists with deduplication.

    This class provides a reusable pattern for managing history lists in settings,
    with configurable validation, normalization, and comparison strategies.
    """

    def __init__(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        settings_key: str,
        max_items: int,
        normalizer: Optional[Callable[[str], str]] = None,
        validator: Optional[Callable[[str], None]] = None,
        comparer: Optional[Callable[[str], str]] = None,
    ) -> None:
        """
        Initialize a history list manager.

        Args:
            settings_key: The key in settings dict where this history is stored
            max_items: Maximum number of items to keep in history
            normalizer: Optional function to normalize items before storage (default: identity)
            validator: Optional function to validate items before storage (called if provided, raises on invalid)
            comparer: Optional function to normalize items for duplicate detection (default: identity)

        """
        self.settings_key = settings_key
        self.max_items = max_items
        self.normalizer = normalizer or (lambda x: x)
        self.validator = validator
        self.comparer = comparer or (lambda x: x)

    def get_items(self, settings: dict[str, Any]) -> list[str]:
        """
        Get the history list from settings.

        Args:
            settings: The settings dictionary

        Returns:
            List of items in most-recent-first order, filtered to valid strings only

        """
        history = settings.get(self.settings_key, [])

        # Handle corrupted data gracefully
        if not isinstance(history, list):
            return []

        # Ensure all entries are strings
        return [item for item in history if isinstance(item, str)]

    def store_item(self, item: str, settings: dict[str, Any]) -> dict[str, Any]:
        """
        Store an item in the history list.

        The item is added to the front of the list. If it already exists,
        it is moved to the front. The list is limited to max_items.

        Performs minimal validation to reject truly invalid inputs (empty strings, null bytes),
        then calls the validator if one was provided during initialization.

        Args:
            item: The item to store
            settings: The settings dictionary

        Returns:
            Updated settings dictionary

        Raises:
            ValueError: If item is empty, contains null byte, or fails validator check

        """
        # Minimal validation: reject truly invalid inputs
        if not item:
            msg = _("Cannot store empty string in history")
            raise ValueError(msg)
        if "\x00" in item:
            msg = _("Cannot store item with null byte in history")
            raise ValueError(msg)

        # Call the validator if provided
        if self.validator is not None:
            self.validator(item)

        history = settings.get(self.settings_key, [])

        # Handle corrupted data
        if not isinstance(history, list):
            history = []

        # Normalize the item for storage
        normalized_item = self.normalizer(item)

        # Remove duplicates using equality comparer
        history = self._remove_duplicate(history, normalized_item)

        # Add to front
        history.insert(0, normalized_item)

        # Limit to max items
        history = history[: self.max_items]

        # Update settings
        settings[self.settings_key] = history
        return settings

    def remove_item(self, item: str, settings: dict[str, Any]) -> dict[str, Any]:
        """
        Remove an item from the history list.

        Args:
            item: The item to remove
            settings: The settings dictionary

        Returns:
            Updated settings dictionary

        """
        history = settings.get(self.settings_key, [])

        if not isinstance(history, list):
            return settings

        # Remove the item
        history = self._remove_duplicate(history, item)

        # Update settings
        settings[self.settings_key] = history
        return settings

    def _remove_duplicate(self, history: list[str], item: str) -> list[str]:
        """
        Remove duplicate item from list while preserving order.

        Args:
            history: List of items
            item: Item to remove if it exists

        Returns:
            List with duplicate removed, order preserved

        """
        normalized_target = self.comparer(item)
        return [h for h in history if self.comparer(h) != normalized_target]



#!/usr/bin/env python3

"""
BDD-style unit tests for path normalization in backend_filesystem_program_settings.py.

This file contains tests for the path normalization logic for the Recent Vehicle
Directories History feature.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2026 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import platform
from unittest.mock import patch

from ardupilot_methodic_configurator.backend_filesystem_program_settings import (
    ProgramSettings,
    normalize_for_comparison,
    normalize_path,
)

# pylint: disable=protected-access


class TestIsTemplateDirectory:
    """Unit tests for _is_template_directory() method."""

    def test_identifies_template_directory(self) -> None:
        """
        Identifies paths within template base directory.

        GIVEN: A path that is within the templates base directory
        WHEN: _is_template_directory is called
        THEN: Should return True
        """
        with patch.object(ProgramSettings, "get_templates_base_dir", return_value="/app/vehicle_templates"):
            # Path inside templates should return True
            result = ProgramSettings._is_template_directory("/app/vehicle_templates/ArduCopter/QuadX")
            assert result is True

    def test_identifies_non_template_directory(self) -> None:
        """
        Identifies paths outside template base directory.

        GIVEN: A path that is NOT within the templates base directory
        WHEN: _is_template_directory is called
        THEN: Should return False
        """
        with patch.object(ProgramSettings, "get_templates_base_dir", return_value="/app/vehicle_templates"):
            # Path outside templates should return False
            result = ProgramSettings._is_template_directory("/home/user/vehicles/my_project")
            assert result is False

    def test_handles_path_resolution_error_conservatively(self) -> None:
        """
        Handles path resolution errors conservatively.

        GIVEN: A path that causes ValueError during commonpath check
        WHEN: _is_template_directory is called
        THEN: Should return True (conservative - don't migrate problematic paths)
        """
        with (
            patch.object(ProgramSettings, "get_templates_base_dir", return_value="/templates"),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path") as mock_os_path,
        ):
            # Make commonpath() raise ValueError (different drives on Windows)
            mock_os_path.normpath.side_effect = lambda x: x
            mock_os_path.abspath.side_effect = lambda x: x
            mock_os_path.commonpath.side_effect = ValueError("Paths on different drives")

            result = ProgramSettings._is_template_directory("/some/path")
            # Should return True conservatively (don't migrate paths that cause errors)
            assert result is True

    def test_handles_os_error_conservatively(self) -> None:
        """
        Handles OS errors conservatively.

        GIVEN: A path that causes OSError during path operations
        WHEN: _is_template_directory is called
        THEN: Should return True (conservative - don't migrate inaccessible paths)
        """
        with (
            patch.object(ProgramSettings, "get_templates_base_dir", return_value="/templates"),
            patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.os_path") as mock_os_path,
        ):
            # Make commonpath() raise OSError
            mock_os_path.normpath.side_effect = lambda x: x
            mock_os_path.abspath.side_effect = lambda x: x
            mock_os_path.commonpath.side_effect = OSError("Network error")

            result = ProgramSettings._is_template_directory("\\\\network\\share\\path")
            # Should return True conservatively
            assert result is True


class TestMigrateLegacyVehicleDir:
    """Unit tests for migrate_settings_to_latest_version() method."""

    def test_migrates_valid_legacy_path(self) -> None:
        """
        Migrates valid legacy vehicle_dir to recent_vehicle_history.

        GIVEN: Settings with Format version 1 and valid legacy vehicle_dir
        WHEN: migrate_settings_to_latest_version is called
        THEN: Legacy path should be migrated to recent_vehicle_history[0]
        AND: Format version should be updated to 2
        """
        test_path = "C:\\vehicles\\old" if platform.system() == "Windows" else "/vehicles/old"
        captured_settings = {}

        def capture(settings: dict) -> None:
            captured_settings.update(settings)

        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get,
            patch.object(ProgramSettings, "_set_settings_from_dict", side_effect=capture) as mock_set,
            patch.object(ProgramSettings, "_is_template_directory", return_value=False),
        ):
            mock_get.return_value = {
                "Format version": 1,
                "recent_vehicle_history": [],
                "directory_selection": {"vehicle_dir": test_path},
            }

            ProgramSettings.migrate_settings_to_latest_version()

            # Should have persisted the migration
            mock_set.assert_called_once()
            assert captured_settings["Format version"] == 2
            assert len(captured_settings["recent_vehicle_history"]) == 1

    def test_migrates_any_legacy_path(self) -> None:
        """
        Migrates any legacy vehicle_dir to history (permissive approach).

        GIVEN: Settings with Format version 1 and legacy vehicle_dir
        WHEN: migrate_settings_to_latest_version is called
        THEN: Path is migrated to history regardless of validity, legacy setting removed
        AND: Format version should be updated to 2
        """
        captured_settings = {}

        def capture(settings: dict) -> None:
            captured_settings.update(settings)

        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get,
            patch.object(ProgramSettings, "_set_settings_from_dict", side_effect=capture) as mock_set,
            patch.object(ProgramSettings, "_is_template_directory", return_value=False),
        ):
            mock_get.return_value = {
                "Format version": 1,
                "recent_vehicle_history": [],
                "directory_selection": {"vehicle_dir": "any/path/here"},
            }

            ProgramSettings.migrate_settings_to_latest_version()

            # Should have persisted changes (removed legacy setting, added to history)
            mock_set.assert_called_once()
            assert captured_settings["Format version"] == 2
            # Legacy setting should be removed
            assert "vehicle_dir" not in captured_settings["directory_selection"]
            # Path should be in history (permissive approach)
            assert len(captured_settings["recent_vehicle_history"]) == 1

    def test_skips_template_paths(self) -> None:
        """
        Removes template paths from legacy settings without migrating.

        GIVEN: Settings with Format version 1 and vehicle_dir pointing to template directory
        WHEN: migrate_settings_to_latest_version is called
        THEN: Template path not added to history, but legacy setting is removed
        AND: Format version should be updated to 2
        """
        captured_settings = {}

        def capture(settings: dict) -> None:
            captured_settings.update(settings)

        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get,
            patch.object(ProgramSettings, "_set_settings_from_dict", side_effect=capture) as mock_set,
            patch.object(ProgramSettings, "_is_template_directory", return_value=True),  # Is template
        ):
            mock_get.return_value = {
                "Format version": 1,
                "recent_vehicle_history": [],
                "directory_selection": {"vehicle_dir": "/templates/ArduCopter/QuadX"},
            }

            ProgramSettings.migrate_settings_to_latest_version()

            # Should have persisted to remove legacy template path
            mock_set.assert_called_once()
            assert captured_settings["Format version"] == 2
            # Legacy setting should be removed
            assert "vehicle_dir" not in captured_settings["directory_selection"]
            # Template path should NOT be in history
            assert len(captured_settings["recent_vehicle_history"]) == 0

    def test_handles_corrupted_history_data(self) -> None:
        """
        Handles corrupted recent_vehicle_history data gracefully.

        GIVEN: Settings with Format version 1 where recent_vehicle_history is not a list (corrupted)
        WHEN: migrate_settings_to_latest_version is called
        THEN: Should treat as empty list and continue migration
        AND: Format version should be updated to 2
        """
        test_path = "C:\\vehicles\\old" if platform.system() == "Windows" else "/vehicles/old"
        captured_settings = {}

        def capture(settings: dict) -> None:
            captured_settings.update(settings)

        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get,
            patch.object(ProgramSettings, "_set_settings_from_dict", side_effect=capture),
            patch.object(ProgramSettings, "_is_template_directory", return_value=False),
        ):
            mock_get.return_value = {
                "Format version": 1,
                "recent_vehicle_history": "corrupted_string",  # Should be list
                "directory_selection": {"vehicle_dir": test_path},
            }

            ProgramSettings.migrate_settings_to_latest_version()

            # Should have treated corrupted data as empty and migrated
            assert captured_settings["Format version"] == 2
            assert isinstance(captured_settings["recent_vehicle_history"], list)

    def test_avoids_duplicate_migration(self) -> None:
        """
        Avoids migrating path that already exists in recent_vehicle_history.

        GIVEN: Settings with Format version 1 where vehicle_dir already exists in recent_vehicle_history
        WHEN: migrate_settings_to_latest_version is called
        THEN: Should not create duplicate entry but still remove legacy setting
        AND: Format version should be updated to 2
        """
        test_path = "C:\\vehicles\\existing" if platform.system() == "Windows" else "/vehicles/existing"
        captured_settings = {}

        def capture(settings: dict) -> None:
            captured_settings.update(settings)

        with (
            patch.object(ProgramSettings, "_get_settings_as_dict") as mock_get,
            patch.object(ProgramSettings, "_set_settings_from_dict", side_effect=capture) as mock_set,
            patch.object(ProgramSettings, "_is_template_directory", return_value=False),
        ):
            mock_get.return_value = {
                "Format version": 1,
                "recent_vehicle_history": [test_path],  # Already in history
                "directory_selection": {"vehicle_dir": test_path},
            }

            ProgramSettings.migrate_settings_to_latest_version()

            # Should have persisted to remove legacy setting even though path already in history
            mock_set.assert_called_once()
            assert captured_settings["Format version"] == 2
            # Legacy setting should be removed
            assert "vehicle_dir" not in captured_settings["directory_selection"]
            # Should NOT have created duplicate - still only 1 entry
            assert len(captured_settings["recent_vehicle_history"]) == 1
            assert captured_settings["recent_vehicle_history"][0] == test_path


class TestPathNormalization:
    """Unit tests for path normalization behavior."""

    def test_normalize_resolves_relative_components(self) -> None:
        """
        Path normalization resolves relative components.

        GIVEN: A path with . and .. components
        WHEN: _normalize_path is called
        THEN: Should resolve relative components
        """
        # This test documents that Path() does MORE than just normalize separators
        with patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.platform_system") as mock_platform:
            mock_platform.return_value = "Linux"
            # Note: Path normalization may resolve these components
            result = normalize_path("/path/./to/../dir")
            # The exact result depends on Path() behavior - just verify it doesn't error
            assert result is not None

    def test_normalize_for_comparison_handles_case_on_windows(self) -> None:
        """
        Normalize for comparison is case-insensitive on Windows.

        GIVEN: Two paths differing only in case
        WHEN: _normalize_for_comparison is called on Windows
        THEN: Should return same normalized form
        """
        with patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.platform_system") as mock_platform:
            mock_platform.return_value = "Windows"
            path1 = normalize_for_comparison("C:\\Path\\To\\Dir")
            path2 = normalize_for_comparison("c:\\path\\to\\dir")
            # On Windows, should normalize to same value (case-insensitive)
            assert path1.lower() == path2.lower()

    def test_normalize_for_comparison_case_sensitive_on_unix(self) -> None:
        """
        Normalize for comparison is case-sensitive on Unix.

        GIVEN: Two paths differing only in case
        WHEN: _normalize_for_comparison is called on Unix
        THEN: May return different normalized forms (case-sensitive)
        """
        with patch("ardupilot_methodic_configurator.backend_filesystem_program_settings.platform_system") as mock_platform:
            mock_platform.return_value = "Linux"
            # On Unix, case matters in paths - but normpath may still affect it
            result1 = normalize_for_comparison("/Path/To/Dir")
            result2 = normalize_for_comparison("/path/to/dir")
            # The results exist and are processed
            assert result1 is not None
            assert result2 is not None



#!/usr/bin/env python3

"""
Behavior-driven tests for FlightControllerConnectionProgress class.

This file is part of ArduPilot Methodic Configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2026 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from unittest.mock import MagicMock, patch

import pytest

from ardupilot_methodic_configurator.frontend_tkinter_flightcontroller_connection_progress import (
    PROGRESS_FC_INIT_COMPLETE,
    FlightControllerConnectionProgress,
)

# pylint: disable=redefined-outer-name, unused-argument


@pytest.fixture
def mock_progress_window() -> tuple[MagicMock, MagicMock]:
    """Fixture providing a mock ProgressWindow for testing."""
    with patch(
        "ardupilot_methodic_configurator.frontend_tkinter_flightcontroller_connection_progress.ProgressWindow"
    ) as mock_pw:
        mock_progress_instance = MagicMock()
        mock_progress_instance.progress_window = MagicMock()
        mock_pw.return_value = mock_progress_instance
        yield mock_pw, mock_progress_instance


@pytest.fixture
def mock_tkinter_root() -> tuple[MagicMock, MagicMock, MagicMock]:
    """Fixture providing a mock Tk root window."""
    with (
        patch("tkinter.Tk") as mock_tk,
        patch("tkinter.ttk.Style") as mock_style,
    ):
        mock_root = MagicMock()
        mock_root.winfo_width.return_value = 1  # Minimal temp root
        mock_tk.return_value = mock_root

        mock_style_instance = MagicMock()
        mock_style.return_value = mock_style_instance

        yield mock_tk, mock_root, mock_style_instance


class TestFlightControllerConnectionProgressBehavior:
    """Test user-visible behavior of flight controller connection progress UI."""

    def test_user_sees_progress_window_during_connection(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        User sees progress window when connecting to flight controller.

        GIVEN: User initiates connection to flight controller
        WHEN: Connection process begins
        THEN: A progress window should be created and displayed
        """
        # Arrange: Mock dependencies
        mock_pw_class, _ = mock_progress_window

        # Act: User triggers flight controller connection
        progress = FlightControllerConnectionProgress()

        # Assert: Progress window should be created
        mock_pw_class.assert_called_once()
        assert progress.progress_window is not None

    def test_connection_window_is_themed_and_positioned(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        Connection window uses proper theme and positioning.

        GIVEN: Connection progress window is created
        WHEN: Window is initialized
        THEN: Should apply theme and center on screen
        """
        # Arrange: Mock dependencies
        _, _, mock_style_instance = mock_tkinter_root

        # Act: Create connection progress
        FlightControllerConnectionProgress()

        # Assert: Theme configured (centering is handled by ProgressWindow)
        mock_style_instance.theme_use.assert_called_once_with("alt")

    def test_user_sees_initialization_progress_updates(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        User sees initialization progress updates during connection.

        GIVEN: Connection to flight controller is in progress
        WHEN: Initialization phase progresses
        THEN: Progress bar should update to reflect current initialization status
        """
        # Arrange: Create progress window
        _, mock_pw_instance = mock_progress_window
        progress = FlightControllerConnectionProgress()

        # Act: Simulate initialization progress updates (0-100%)
        progress.update_init_progress_bar(0, 100)
        progress.update_init_progress_bar(50, 100)
        progress.update_init_progress_bar(100, 100)

        # Assert: Progress should be mapped to 0-20% range (PROGRESS_FC_INIT_COMPLETE)
        calls = mock_pw_instance.update_progress_bar.call_args_list
        assert len(calls) == 3
        assert calls[0][0] == (0, 100)  # 0% -> 0
        assert calls[1][0] == (10, 100)  # 50% -> 10 (50% of 20)
        assert calls[2][0] == (20, 100)  # 100% -> 20

    def test_user_sees_connection_progress_updates(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        User sees connection progress updates during connection phase.

        GIVEN: Flight controller initialization is complete
        WHEN: Connection phase progresses
        THEN: Progress bar should update to reflect current connection status
        """
        # Arrange: Create progress window
        _, mock_pw_instance = mock_progress_window
        progress = FlightControllerConnectionProgress()

        # Act: Simulate connection progress updates (0-100%)
        progress.update_connect_progress_bar(0, 100)
        progress.update_connect_progress_bar(50, 100)
        progress.update_connect_progress_bar(100, 100)

        # Assert: Progress should be mapped to 20-100% range
        calls = mock_pw_instance.update_progress_bar.call_args_list
        assert len(calls) == 3
        assert calls[0][0] == (20, 100)  # 0% -> 20 (PROGRESS_FC_INIT_COMPLETE)
        assert calls[1][0] == (60, 100)  # 50% -> 60 (20 + 50% of 80)
        assert calls[2][0] == (100, 100)  # 100% -> 100

    def test_user_can_use_connection_progress_as_context_manager(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        User benefits from automatic cleanup when using connection progress.

        GIVEN: Connection progress is used in a context manager
        WHEN: Context exits (normally or with exception)
        THEN: Resources should be cleaned up automatically
        """
        # Arrange: Mock dependencies
        _, mock_root, _ = mock_tkinter_root
        _, mock_pw_instance = mock_progress_window

        # Act: Use as context manager
        with FlightControllerConnectionProgress() as progress:
            assert progress is not None
            # Simulate some work
            progress.update_init_progress_bar(50, 100)

        # Assert: Cleanup should have been called
        mock_pw_instance.destroy.assert_called_once()
        mock_root.destroy.assert_called_once()

    def test_user_informed_of_invalid_progress_values(self, mock_tkinter_root, mock_progress_window, caplog) -> None:
        """
        User is protected from invalid progress values.

        GIVEN: Progress update with invalid max_value
        WHEN: Application tries to update progress with invalid data
        THEN: Should log error and not crash
        AND: Should not attempt to update progress bar
        """
        # Arrange: Create progress window
        _, mock_pw_instance = mock_progress_window
        progress = FlightControllerConnectionProgress()
        initial_call_count = mock_pw_instance.update_progress_bar.call_count

        # Act: Try to update with invalid max_value
        with caplog.at_level("ERROR"):
            progress.update_init_progress_bar(50, 0)  # Invalid: max_value=0
            progress.update_connect_progress_bar(50, -1)  # Invalid: max_value=-1

        # Assert: Should log error and not call update_progress_bar
        assert len(caplog.records) == 2
        assert "Invalid max_value" in caplog.text
        assert mock_pw_instance.update_progress_bar.call_count == initial_call_count

    def test_progress_calculation_boundary_conditions(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        Progress calculations handle boundary conditions correctly.

        GIVEN: Various boundary conditions for progress values
        WHEN: Progress is updated with edge case values
        THEN: Calculations should handle boundaries gracefully
        """
        # Arrange: Create progress window
        _, mock_pw_instance = mock_progress_window
        progress = FlightControllerConnectionProgress()

        # Act & Assert: Test initialization phase boundaries
        progress.update_init_progress_bar(0, 100)
        assert mock_pw_instance.update_progress_bar.call_args[0] == (0, 100)

        progress.update_init_progress_bar(100, 100)
        assert mock_pw_instance.update_progress_bar.call_args[0] == (
            PROGRESS_FC_INIT_COMPLETE,
            100,
        )

        # Act & Assert: Test connection phase boundaries
        progress.update_connect_progress_bar(0, 100)
        assert mock_pw_instance.update_progress_bar.call_args[0] == (
            PROGRESS_FC_INIT_COMPLETE,
            100,
        )

        progress.update_connect_progress_bar(100, 100)
        assert mock_pw_instance.update_progress_bar.call_args[0] == (100, 100)

    def test_cleanup_handles_missing_resources_gracefully(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        Cleanup handles missing or already-destroyed resources gracefully.

        GIVEN: Progress window with potentially missing resources
        WHEN: Destroy is called
        THEN: Should not raise exceptions even if resources are missing
        """
        # Arrange: Create progress window
        progress = FlightControllerConnectionProgress()

        # Simulate missing progress_window (using type: ignore since type is non-Optional)
        progress.progress_window = None  # type: ignore[assignment]

        # Act & Assert: Should not raise exception
        progress.destroy()

        # Simulate missing temp_root
        progress.temp_root = None  # type: ignore[assignment]
        progress.destroy()  # Should still not raise

    def test_context_manager_cleanup_on_exception(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        Context manager ensures cleanup even when exception occurs.

        GIVEN: Connection progress used in context manager
        WHEN: Exception occurs during context
        THEN: Cleanup should still happen
        AND: Exception should propagate
        """
        # Arrange: Mock dependencies
        _, mock_root, _ = mock_tkinter_root
        _, mock_pw_instance = mock_progress_window

        # Act: Exception should propagate but cleanup should happen
        error_msg = "Test exception"
        with pytest.raises(ValueError, match=error_msg), FlightControllerConnectionProgress():
            raise ValueError(error_msg)

        # Assert: Cleanup should have been called despite exception
        mock_pw_instance.destroy.assert_called_once()
        mock_root.destroy.assert_called_once()


class TestProgressMappingAccuracy:
    """Test progress value mapping accuracy."""

    def test_init_progress_maps_correctly_across_range(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        Initialization progress maps correctly across entire range.

        GIVEN: Initialization progress values from 0-100%
        WHEN: Each value is mapped to display range
        THEN: Should map proportionally to 0-20% range
        """
        # Arrange
        _, mock_pw_instance = mock_progress_window
        progress = FlightControllerConnectionProgress()

        test_cases = [
            (0, 0),
            (10, 2),
            (25, 5),
            (50, 10),
            (75, 15),
            (100, 20),
        ]

        # Act & Assert: Test each mapping
        for input_val, expected_output in test_cases:
            progress.update_init_progress_bar(input_val, 100)
            actual_output = mock_pw_instance.update_progress_bar.call_args[0][0]
            assert actual_output == expected_output, f"Failed for input {input_val}"

    def test_connect_progress_maps_correctly_across_range(self, mock_tkinter_root, mock_progress_window) -> None:
        """
        Connection progress maps correctly across entire range.

        GIVEN: Connection progress values from 0-100%
        WHEN: Each value is mapped to display range
        THEN: Should map proportionally to 20-100% range
        """
        # Arrange
        _, mock_pw_instance = mock_progress_window
        progress = FlightControllerConnectionProgress()

        test_cases = [
            (0, 20),
            (10, 28),
            (25, 40),
            (50, 60),
            (75, 80),
            (100, 100),
        ]

        # Act & Assert: Test each mapping
        for input_val, expected_output in test_cases:
            progress.update_connect_progress_bar(input_val, 100)
            actual_output = mock_pw_instance.update_progress_bar.call_args[0][0]
            assert actual_output == expected_output, f"Failed for input {input_val}"
